package ru.dbotthepony.kstarbound.world.entities

import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.gson.getArray
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.io.IntValueCodec
import ru.dbotthepony.kommons.io.map
import ru.dbotthepony.kommons.matrix.Matrix3f
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.animation.AnimatedPartsDefinition
import ru.dbotthepony.kstarbound.defs.animation.AnimationDefinition
import ru.dbotthepony.kstarbound.defs.animation.ParticleFactory
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.math.PeriodicFunction
import ru.dbotthepony.kstarbound.math.approachAngle
import ru.dbotthepony.kstarbound.network.syncher.InternedStringCodec
import ru.dbotthepony.kstarbound.network.syncher.NetworkedElement
import ru.dbotthepony.kstarbound.network.syncher.NetworkedGroup
import ru.dbotthepony.kstarbound.network.syncher.NetworkedList
import ru.dbotthepony.kstarbound.network.syncher.NetworkedMap
import ru.dbotthepony.kstarbound.network.syncher.NetworkedSignal
import ru.dbotthepony.kstarbound.network.syncher.networkedAABBNullable
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedColor
import ru.dbotthepony.kstarbound.network.syncher.networkedEventCounter
import ru.dbotthepony.kstarbound.network.syncher.networkedFixedPoint
import ru.dbotthepony.kstarbound.network.syncher.networkedFloat
import ru.dbotthepony.kstarbound.network.syncher.networkedPointer
import ru.dbotthepony.kstarbound.network.syncher.networkedSignedInt
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import ru.dbotthepony.kstarbound.network.syncher.networkedUnsignedInt
import ru.dbotthepony.kstarbound.util.random.random
import java.util.*
import java.util.function.Consumer
import kotlin.math.atan2
import kotlin.math.cos
import kotlin.math.sin
import kotlin.math.sqrt

// NetworkedAnimator + AnimatedPartSet combined into one class

// AnimatedPartSet original docs:
// Defines a "animated" data set constructed in such a way that it is very
// useful for doing generic animations with lots of additional animation data.
// It is made up of two concepts, "states" and "parts".
//
// States:
//
// There are N "state types" defined, which each defines a set of mutually
// exclusive states that each "state type" can be in.  For example, one state
// type might be "movement", and the "movement" states might be "idle", "walk",
// and "run.  Another state type might be "attack" which could have as its
// states "idle", and "melee".  Each state type will have exactly one currently
// active state, so this class may, for example,  be in the total state of
// "movement:idle" and "attack:melee".  Each state within each state type is
// animated, so that over time the state frame increases and may loop around,
// or transition into another state so that that state type without interaction
// may go from "melee" to "idle" when the "melee" state animation is finished.
// This is defined by the individual state config in the configuration passed
// into the constructor.
//
// Parts:
//
// Each instance of this class also can have N "Parts" defined, which are
// groups of properties that "listen" to active states.  Each part can "listen"
// to one or more state types, and the first matching state x state type pair
// (in order of state type priority which is specified in the config) is
// chosen, and the properties from that state type and state are merged into
// the part to produce the final active part information.  Rather than having a
// single image or image set for each part, since this class is intended to be
// as generic as possible, all of this data is assumed to be queried from the
// part properties, so that things such as image data as well as other things
// like damage or collision polys can be stored along with the animation
// frames, the part state, the base part, whichever is most applicable.
class Animator() {
	private class Light {
		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var active by networkedBoolean().also { elements.add(it) }
		var xPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var color by networkedColor().also { elements.add(it) }
		var pointAngle by networkedFixedPoint(0.01).also { elements.add(it); it.interpolator = Interpolator.Linear }

		var anchorPart: String? = null
		var transformationGroups: List<String> = listOf()
		var rotationGroup: String? = null
		var rotationCenter: Vector2d? = null

		var flicker: PeriodicFunction? = null
		var pointLight: Boolean = false
		var pointBeam: Float = 0f
		var beamAmbience: Float = 0f
	}

	private enum class SoundSignal {
		PLAY, STOP_ALL;

		companion object {
			val CODEC = IntValueCodec.map({ entries[this] }, { ordinal })
		}
	}

	private class Sound {
		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var rangeMultiplier = 0.0

		val soundPool = NetworkedList(InternedStringCodec).also { elements.add(it) }
		var xPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yPosition by networkedFixedPoint(0.0125).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var volumeTarget by networkedFloat(1.0).also { elements.add(it) }
		var volumeRampTime by networkedFloat(0.0).also { elements.add(it) }
		var pitchMultiplierTarget by networkedFloat(1.0).also { elements.add(it) }
		var pitchMultiplierRampTime by networkedFloat(0.0).also { elements.add(it) }
		var loops by networkedSignedInt().also { elements.add(it) }
		val signals = NetworkedSignal(SoundSignal.CODEC).also { elements.add(it) }
	}

	private class Effect(val type: String, val time: Double, val directives: String) {
		val enabled = networkedBoolean()
		var timer: Double = 0.0

		fun tick(delta: Double) {
			if (timer <= 0.0)
				timer = time
			else
				timer -= delta
		}
	}

	private class StateType(config: AnimatedPartsDefinition.StateType) {
		// NetworkedAnimator
		private var noPropagate = false
		val stateIndex = networkedPointer(-1L)
		val startedEvent = networkedEventCounter()

		// AnimatedPartSet
		var enabled = config.enabled
		var activeStateDirty = true
		val priority = config.priority
		val stateTypeProperties = config.properties
		val default: String
		var activeProperties = JsonObject()
			private set

		// sorted by key
		// Basically, this is definition of each separate state
		val states = config.states

		var timer = 0.0
		var frame = 0
			private set(value) {
				if (field != value) {
					field = value
					frameChanged = true
				}
			}

		init {
			if (states.isNotEmpty() && config.default.isBlank())
				default = states.keys.first()
			else
				default = config.default
		}

		private var activeStateChanged = false
		private var frameChanged = false

		var activeState: AnimatedPartsDefinition.StateType.State? = null
			set(value) {
				activeStateChanged = true

				if (value == null) {
					if (!noPropagate) {
						try {
							noPropagate = true
							stateIndex.accept(-1L)
						} finally {
							noPropagate = false
						}
					}

					field = null
				} else {
					if (!noPropagate) {
						try {
							noPropagate = true
							stateIndex.accept(value.index.toLong())
						} finally {
							noPropagate = false
						}
					}

					field = value
				}
			}

		fun set(state: String, alwaysStart: Boolean): Boolean {
			val getState = states[state] ?: return false

			if (activeState != getState || alwaysStart) {
				activeState = getState
				timer = 0.0
				return true
			}

			return false
		}

		init {
			if (default in states) {
				activeState = states[default]!!
				stateIndex.accept(activeState!!.index.toLong())
			}

			stateIndex.addListener(Consumer {
				if (noPropagate) return@Consumer

				if (it == -1L) {
					try {
						noPropagate = true
						activeState = null
					} finally {
						noPropagate = false
					}
				} else {
					try {
						noPropagate = true
						set(states.keys.elementAtOrNull(it.toInt()) ?: throw IllegalArgumentException("Unknown animation state $it!"), true)
					} finally {
						noPropagate = false
					}
				}
			})
		}

		fun tick(delta: Double) {
			var activeState = activeState ?: return
			timer += delta

			if (timer > activeState.cycle) {
				when (activeState.mode) {
					AnimatedPartsDefinition.AnimationMode.END -> timer = activeState.cycle

					AnimatedPartsDefinition.AnimationMode.TRANSITION -> {
						activeState = states[activeState.transition]!! // validity of 'transition' is checked during json load
						this.activeState = activeState
						timer = 0.0
					}

					AnimatedPartsDefinition.AnimationMode.LOOP -> timer %= activeState.cycle
				}
			}

			frame = (timer / activeState.cycle * activeState.frames).toInt().coerceIn(0, activeState.frames - 1)

			if (activeStateChanged || frameChanged) {
				activeProperties = mergeJson(stateTypeProperties.deepCopy(), activeState.properties)
				frameChanged = false
				activeStateChanged = false

				for ((key, values) in activeState.frameProperties) {
					if (values.size() >= frame) {
						activeProperties[key] = values[frame].deepCopy()
					}
				}
			}
		}
	}

	private class Part(config: AnimatedPartsDefinition.Part) {
		val partProperties = config.properties
		var activePartDirty = true
		val partStates = config.partStates
	}

	private class RotationGroup {
		var angularVelocity = 0.0
		var rotationCenter = Vector2d.ZERO
		val targetAngle = networkedFloat()
		var currentAngle = 0.0
		val immediateEvent = networkedEventCounter()

		fun tick(delta: Double) {
			if (angularVelocity == 0.0) {
				currentAngle = targetAngle.get()
			} else {
				currentAngle = approachAngle(targetAngle.get(), currentAngle, angularVelocity * delta)
			}
		}
	}

	private class TransformationGroup {
		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var interpolated = false

		var xTranslation by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yTranslation by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var xScale by networkedFloat(1.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yScale by networkedFloat(1.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var xShear by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }
		var yShear by networkedFloat(0.0).also { elements.add(it); it.interpolator = Interpolator.Linear }

		fun affineTransform(): Matrix3f {
			return Matrix3f.rowMajor(
				(xScale * cos(xShear)).toFloat(), (xScale * sin(xShear)).toFloat(), xTranslation.toFloat(),
				(yScale * sin(yShear)).toFloat(), (yScale * cos(yShear)).toFloat(), yTranslation.toFloat(),
				0f, 0f, 1f
			)
		}

		fun setAffineTransform(value: Matrix3f) {
			xTranslation = value[2, 0].toDouble()
			yTranslation = value[2, 1].toDouble()

			xScale = sqrt(value[0, 0].toDouble() * value[0, 0].toDouble() + value[1, 0].toDouble() * value[1, 0].toDouble())
			yScale = sqrt(value[0, 1].toDouble() * value[0, 1].toDouble() + value[1, 1].toDouble() * value[1, 1].toDouble())

			xShear = atan2(value[1, 0].toDouble(), value[0, 0].toDouble())
			yShear = atan2(value[0, 1].toDouble(), value[1, 1].toDouble())
		}
	}

	private class ParticleEmitter {
		data class Config(val count: Int, val offset: Vector2d, val flip: Boolean, val factory: ParticleFactory)

		private val elements = ArrayList<NetworkedElement>()

		fun addTo(group: NetworkedGroup) {
			elements.forEach { group.add(it) }
		}

		var emissionRate by networkedFloat().also { elements.add(it) }
		var burstCount by networkedUnsignedInt().also { elements.add(it) }
		var randomSelectCount by networkedUnsignedInt().also { elements.add(it) }
		var emissionRateVariance = 0.0
		var offsetRegion by networkedAABBNullable().also { elements.add(it) }
		var anchorPart: String? = null
		var transformationGroups: List<String> = listOf()
		var rotationGroup: String? = null
		var rotationCenter: Vector2d? = null

		val particleList = ArrayList<Config>()

		var active by networkedBoolean().also { elements.add(it) }
		val burstEvent = networkedEventCounter().also { elements.add(it); it.ignoreOccurrencesOnLoad = true }

		var timer = 0.0
	}

	val networkGroup = NetworkedGroup()

	private val elements = ArrayList<NetworkedElement>()

	var processingDirectives by networkedString().also { elements.add(it) }
	var zoom by networkedFloat().also { elements.add(it) }
	var isFlipped by networkedBoolean().also { elements.add(it) }
	var flippedRelativeCenterLine by networkedFloat().also { elements.add(it) }
	var animationRate by networkedFloat(1.0).also { elements.add(it); it.interpolator = Interpolator.Linear }

	private val globalTags = NetworkedMap(InternedStringCodec, InternedStringCodec)
	private val parts = Object2ObjectAVLTreeMap<String, Part>()
	private val partTags = HashMap<String, NetworkedMap<String, String>>()

	// iterate by priority, network by sorted key
	private val stateTypes = LinkedHashMap<String, StateType>()
	private val rotationGroups = Object2ObjectAVLTreeMap<String, RotationGroup>()
	private val transformationGroups = Object2ObjectAVLTreeMap<String, TransformationGroup>()
	private val particleEmitters = Object2ObjectAVLTreeMap<String, ParticleEmitter>()
	private val lights = Object2ObjectAVLTreeMap<String, Light>()
	private val sounds = Object2ObjectAVLTreeMap<String, Sound>()
	private val effects = Object2ObjectAVLTreeMap<String, Effect>()

	private val random = random()

	init {
		setupNetworkElements()
	}

	constructor(config: AnimationDefinition) : this() {
		for ((k, v) in config.globalTagDefaults) {
			globalTags[k] = v
		}

		for ((part, tags) in config.partTagDefaults) {
			for ((k, v) in tags) {
				setPartTag(part, k, v)
			}
		}

		for ((k, v) in config.transformationGroups) {
			val group = TransformationGroup()
			transformationGroups[k] = group
			group.interpolated = v.interpolated
		}

		for ((k, v) in config.rotationGroups) {
			val group = RotationGroup()
			group.angularVelocity = v.angularVelocity
			group.rotationCenter = v.rotationCenter
			rotationGroups[k] = group
		}

		for ((k, v) in config.particleEmitters) {
			val emitter = ParticleEmitter()

			emitter.emissionRate = v.emissionRate
			emitter.emissionRateVariance = v.emissionRateVariance
			emitter.offsetRegion = KOptional.ofNullable(v.offsetRegion)
			emitter.anchorPart = v.anchorPart
			emitter.transformationGroups = v.transformationGroups
			emitter.rotationGroup = v.rotationGroup
			emitter.rotationCenter = v.rotationCenter
			emitter.burstCount = v.burstCount
			emitter.randomSelectCount = v.randomSelectCount
			emitter.active = v.active

			for (particle in v.particles) {
				val factory = particle.particle.map({ it.value }, { it }) ?: continue // not a valid particle, too bad.
				emitter.particleList.add(ParticleEmitter.Config(particle.count, particle.offset, particle.flip, factory))
			}

			particleEmitters[k] = emitter
		}

		for ((k, v) in config.lights) {
			val light = Light()

			light.active = v.active
			light.xPosition = v.position.x
			light.yPosition = v.position.y
			light.color = v.color
			light.anchorPart = v.anchorPart
			light.transformationGroups = v.transformationGroups
			light.rotationGroup = v.rotationGroup
			light.rotationCenter = v.rotationCenter
			light.pointAngle = Math.toRadians(v.pointAngle)
			light.pointLight = v.pointLight
			light.pointBeam = v.pointBeam
			light.beamAmbience = v.beamAmbience

			if (v.flickerPeriod != null)
				light.flicker = PeriodicFunction(v.flickerPeriod, v.flickerMinIntensity, v.flickerMaxIntensity, v.flickerPeriodVariance, v.flickerIntensityVariance)

			lights[k] = light
		}

		for ((k, v) in config.sounds) {
			val sound = Sound()

			if (v.isLeft) {
				sound.soundPool.addAll(v.left())
			} else {
				val conf = v.right()
				sound.soundPool.addAll(conf.pool.map { it.fullPath })
				sound.xPosition = conf.position.x
				sound.yPosition = conf.position.y
				sound.volumeTarget = conf.volume
				sound.volumeRampTime = conf.volumeRampTime
				sound.pitchMultiplierTarget = conf.pitchMultiplier
				sound.pitchMultiplierRampTime = conf.pitchMultiplierRampTime
				sound.rangeMultiplier = conf.rangeMultiplier
			}

			sounds[k] = sound
		}

		for ((k, v) in config.effects) {
			effects[k] = Effect(v.type, v.time, v.directives)
		}

		if (config.animatedParts != null) {
			for ((k, v) in config.animatedParts.stateTypes.entries.sortedWith { o1, o2 -> o2.value.priority.compareTo(o1.value.priority) }) {
				stateTypes[k] = StateType(v)
			}

			for ((k, v) in config.animatedParts.parts) {
				parts[k] = Part(v)
			}
		}

		for (k in parts.keys) {
			partTags.computeIfAbsent(k) { NetworkedMap(InternedStringCodec, InternedStringCodec) }
		}

		setupNetworkElements()
	}

	// Every part image can have one or more <tag> directives in it, which if set
	// here will be replaced by the tag value when constructing Drawables.  All
	// Drawables can also have a <frame> tag which will be set to whatever the
	// current state frame is (1 indexed, so the first frame is 1).
	fun setGlobalTag(key: String, value: String) {
		globalTags[key] = value
	}

	fun setPartTag(partName: String, tagKey: String, tagValue: String) {
		var tags = partTags[partName]

		if (tags == null) {
			tags = NetworkedMap(InternedStringCodec, InternedStringCodec)
			partTags[partName] = tags
		}

		tags[tagKey] = tagValue
	}

	private fun setupNetworkElements() {
		networkGroup.clear()
		elements.forEach { networkGroup.add(it) }

		networkGroup.add(globalTags)

		// animated part set
		for (v in parts.keys) {
			networkGroup.add(partTags[v] ?: throw RuntimeException("Missing animated part $v!"))
		}

		stateTypes.entries.stream()
			.sorted { o1, o2 -> o1.key.compareTo(o2.key) }
			.map { it.value }
		.forEach {
			networkGroup.add(it.stateIndex)
			networkGroup.add(it.startedEvent)
		}

		for (v in transformationGroups.values) {
			v.addTo(networkGroup)
		}

		for (v in rotationGroups.values) {
			networkGroup.add(v.targetAngle)
			networkGroup.add(v.immediateEvent)
		}

		for (v in particleEmitters.values) {
			v.addTo(networkGroup)
		}

		for (v in lights.values) {
			v.addTo(networkGroup)
		}

		for (v in sounds.values) {
			v.addTo(networkGroup)
		}

		for (v in effects.values) {
			networkGroup.add(v.enabled)
		}
	}

	fun setActiveState(type: String, state: String, alwaysStart: Boolean = false): Boolean {
		val getType = stateTypes[type] ?: return false
		val getState = getType.states[state] ?: return false

		if (getType.activeState != getState || alwaysStart) {
			getType.timer = 0.0
			return true
		}

		return false
	}

	// TODO: Dynamic target
	@Suppress("Name_Shadowing")
	fun tick(delta: Double) {
		val delta = delta * animationRate

		for (state in stateTypes.values) {
			state.tick(delta)

			if ("lightsOn" in state.activeProperties) {
				for (name in state.activeProperties.getArray("lightsOn")) {
					lights[name.asString]?.active = true
				}
			}

			if ("lightsOff" in state.activeProperties) {
				for (name in state.activeProperties.getArray("lightsOff")) {
					lights[name.asString]?.active = false
				}
			}

			if ("particleEmittersOn" in state.activeProperties) {
				for (name in state.activeProperties.getArray("particleEmittersOn")) {
					particleEmitters[name.asString]?.active = true
				}
			}

			if ("particleEmittersOff" in state.activeProperties) {
				for (name in state.activeProperties.getArray("particleEmittersOff")) {
					particleEmitters[name.asString]?.active = false
				}
			}
		}

		for (rotationGroup in rotationGroups.values) {
			rotationGroup.tick(delta)
		}

		for (light in lights.values) {
			light.flicker?.update(delta, random)
		}

		for (effect in effects.values) {
			effect.tick(delta)
		}
	}

	companion object {
		// lame
		fun load(path: String): Animator {
			val json = Starbound.loadJsonAsset(path)

			if (json == null) {
				if (missing.add(path)) {
					LOGGER.error("Unable to instance Animator from $path! This very likely gonna make networking code go haywire.")
				}

				return Animator()
			}

			return Animator(Starbound.gson.fromJson(json, AnimationDefinition::class.java))
		}

		private val LOGGER = LogManager.getLogger()
		private val missing = Collections.synchronizedSet(ObjectOpenHashSet<String>())
	}
}
